/*-
 * Copyright (C) @BABOLO  2002 Dec 23
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2002 Dec 23\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: parser.c,v 1.4 2004/11/06 22:38:00 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <babolo/BLINflag.h>
#include <mife.h>
#include <err.h>
#include <libpq-fe.h>
#include "pgoblin.h"

static const char *cmdstr[] =
{ ""
, "copyin "
, "copyout "
, "binselect "
, "strselect "
, "perform "
, "!"
, NULL
};

static const u_int32_t cmdkind[] =
{ 0
, TEXTPARM
, TEXTPARM
, TEXTPARM
, TEXTPARM
, TEXTPARM
, TEXTPARM
, 0
};

static struct pgoblin *
pgoblin_init() {
    struct pgoblin *pgm;
    struct code_t *ppole;

    ppole = malloc(sizeof(struct code_t) * BUFSIZE);
    pgm = malloc(sizeof(struct pgoblin));
    pgm->maxx = 0;
    pgm->code = ppole;
    return(pgm);
}

static pnode
pgoblin_new(struct pgoblin *pgm) {
    if  (pgm->maxx >= BUFSIZE) errx(EX_SOFTWARE, "Programm too big");
    return(pgm->maxx++);
}

static enum command
pgoblin_seach(u_int32_t flag, char **q) {
    enum command cmd, i = pgob_null;
    size_t l = 0;

    ifBLIN_QV3(flag) fprintf(stderr, "+pgoblin_seach ~%s~\n", *q);
    while (q && *q && **q == ' ') (*q)++;
    for (cmd = pgob_null; cmdstr[cmd]; cmd++) {
        if  (l < strlen(cmdstr[cmd]) && 0 == strncmp(*q, cmdstr[cmd], strlen(cmdstr[cmd]))) {
            l = strlen(cmdstr[cmd]);
            i = cmd;
    }   }
    if  (cmdstr[i]) *q += strlen(cmdstr[i]);
    ifBLIN_QV3(flag) fprintf(stderr, "-pgoblin_seach %d ~%s~\n", i, *q);
    return(i);
}

struct pgoblin *
pgoblin_parser(struct pgoblincon *options, struct mife_descriptor *in) {
    struct pgoblin *pgm;
    enum command cmd;
    char *delim, *q;
    pnode ci;

    ifBLIN_QV3(options->flag) fprintf(stderr, "+pgoblin_parser in=%s\n", in ? "OK" : "NULL");
    if  (pgoblin_shurecon(options)) return(NULL);
    if  (mife_read(in, 0, 0, NULL) <= 0 || !(q = mife_get(in, 0))) {
        ifBLIN_QV1(options->flag) warnx("pgoblin_parser: No program");
        return(NULL);
    }

    delim = "#";
    pgm = pgoblin_init();
    while (*q) {
        ifBLIN_QV3(options->flag) fprintf(stderr, "Next cmd~%s~\n", q);
        if  (0 != strncmp(delim, q, strlen(delim))) {
            ifBLIN_QV1(options->flag) warnx("pgoblin_parser: No command");
            return(NULL);
        }
        q += strlen(delim);
        cmd = pgoblin_seach(options->flag, &q);
        if  (cmd == pgob_illegal) {
            ifBLIN_QV1(options->flag) errx(EX_DATAERR, "syntax");
            exit(EX_DATAERR);
        }
        ci = pgoblin_new(pgm);
        pgm->code[ci].cmd = cmd;
        while (*q && *q != '\n') q++;
        if  (*q) *q++ = '\0';
        ifBLIN_QV3(options->flag)
            fprintf(stderr, "cmd=%d(%s) %08X\n", cmd, cmdstr[cmd], cmdkind[cmd]);
        if  (cmdkind[cmd] & TEXTPARM) {
            int i = 0;
            pgm->code[ci].parm = q;
            while (*q && 0 != strncmp(delim, q, strlen(delim))) {
                ifBLIN_QV4(options->flag) fprintf(stderr, "Next line~%s~\n", q);
                while (*q && *q != '\n') q++;
                q++;
                i = 1;
            }
            if  (i) *(q - 1) = '\0';
              else pgm->code[ci].parm = "";
        } else {
            pgm->code[ci].parm = NULL;
        }
    }
    return(pgm);
}
